"""Deals with CPU related stuffs."""
import psutil


class CPUManager:
    """A class to manage and retrieve CPU usage information."""

    @staticmethod
    def get_cpu_stats() -> tuple:
        """Retrieves the overall current CPU usage percentage and CPU usage by system processes.

        Returns:
            tuple: A tuple containing (overall CPU usage, system process CPU usage).

        Raises:
            None
        """
        return CPUManager._get_cpu_usage()

    @staticmethod
    def _get_cpu_usage() -> tuple:
        """Retrieves the overall current CPU usage percentage.

        Returns:
            tuple: (overall CPU usage percentage, CPU usage by system processes,
                    CPU usage by the current user).

        Raises:
            None
        """
        overall_cpu = psutil.cpu_percent(interval=1)
        cpu_times = psutil.cpu_times_percent(interval=1)
        system_cpu = cpu_times.system
        user_cpu = cpu_times.user

        return overall_cpu, system_cpu, user_cpu



import os
import platform
import socket
from pathlib import Path

IS_WINDOWS_OS = platform.system() == "Windows"


class Config:
    """Static class container for all variables."""

    class DIRECTORIES:
        """Container for any directories required for this library.

        Folders will be created automatically
        """

        OUTPUT = Path().cwd() / "output"
        SCREENSHOTS = os.path.join(OUTPUT, "screenshots")

    if IS_WINDOWS_OS:
        username = os.getlogin()
        hostname = socket.gethostname()
    else:
        username = ""
        hostname = ""


CONFIG = Config()



"""Module to deal with google sheets."""

from json.decoder import JSONDecodeError
from typing import Any

import gspread
from gspread.exceptions import APIError
from retry import retry

from t_desktop.utils.logger import logger


class GSheets:
    """Class to deal with Google sheets rows."""

    def __init__(self, credentials: Any):
        """Initializes the Mapping class."""
        self.client = gspread.authorize(credentials)

    @retry(
        exceptions=(APIError, JSONDecodeError),
        tries=10,
        delay=5,
        backoff=2,
        max_delay=64,
        jitter=(0, 20),
        logger=logger,
    )
    def append_row(self, spreadsheet_id: str, tab: str, row: list) -> None:
        """Appends a row to the end of the specified worksheet.

        Args:
            tab (Any): The tab configuration to access the worksheet.
            row (list): The row data to append.

        Returns:
            None

        Raises:
            ValueError: If the row data is not in list format.
        """
        if not isinstance(row, list):
            raise ValueError("Row data must be provided as a list.")

        spreadsheet = self.client.open_by_key(spreadsheet_id)
        worksheet = spreadsheet.get_worksheet_by_id(tab)

        worksheet.append_row(row, value_input_option="USER_ENTERED")
        logger.info(f"Row appended to {tab}: {row}")



"""Deal with disk related stuffs."""
import psutil


class DiskManager:
    """A class to manage and retrieve disk space information."""

    @staticmethod
    def get_disk_stats() -> tuple[float, float, float]:
        """Retrieves disk space statistics for the system.

        Returns:
            tuple: A tuple containing:
                - Total disk space in gigabytes (GB).
                - Available disk space in gigabytes (GB).
                - Used disk space in gigabytes (GB).

        Raises:
            None
        """
        total, _, free, _ = psutil.disk_usage("/")
        gb_conversion = 1024**3
        total_gb = round(total / gb_conversion, 2)
        free_gb = round(free / gb_conversion, 2)
        used_gb = round(total_gb - free_gb, 2)

        return total_gb, free_gb, used_gb



class UsernameNotFoundError(Exception):
    """Operational System Username not found in env variables."""

    pass


class AppCrashException(Exception):
    """Application Crashed."""

    pass


class LoginMethodNotFoundError(Exception):
    """Custom exception raised when the login method is not found."""

    pass



"""Main module."""
import contextlib
import os
import subprocess
import time
from abc import ABC
from typing import Union
import _ctypes


from t_desktop.config import CONFIG, IS_WINDOWS_OS
from t_desktop.decorators import retry_if_pywin_error
from t_desktop.exceptions import UsernameNotFoundError
from t_desktop.utils.logger import logger

if IS_WINDOWS_OS:
    from pywinauto import Application, WindowSpecification, mouse
    from pywinauto.application import ProcessNotFoundError
    from pywinauto.timings import TimeoutError as PywinTimeout
    from pywinauto.findwindows import ElementNotFoundError
else:
    logger.warning("Not in a Windows Machine, failed to import UI related libraries")


class DesktopApp(ABC):
    """This class is the base class for all Desktop classes."""

    def __init__(self, app_path: str):
        """Initializes the Apps class."""
        self.app: Application = None
        self.app_path = app_path

    @retry_if_pywin_error(retries=3)  # type: ignore[misc]
    def _dialog(self) -> WindowSpecification:
        try:
            return self.app.top_window()
        except RuntimeError as e:
            if (
                "No windows for that process could be found" in str(e)
                and self.get_app_session_if_running(self.app_path) is not None
            ):
                time.sleep(5)
                self.app = self.connect_to_app(self.app_path)
                return self.app.top_window()
            raise e

    @property
    def dialog(self) -> WindowSpecification:
        """The main point of interaction with the App Window."""
        return self._dialog()

    def set_focus(self) -> None:
        """Ste focus to 'self.dialog'."""
        with contextlib.suppress(_ctypes.COMError):
            self.dialog.set_focus()

    @retry_if_pywin_error(exceptions=(OSError, RuntimeError), delay=3, retries=2)  # type: ignore[misc]
    def close_alerts(
        self, expected_window_title: str, max_attempts: int = 3, sleep_seconds: int = 1, ignored_titles: list[str] = []
    ) -> None:
        """Close alerts until the expected window exists or the timeout is reached.

        Args:
            expected_window_title (str): The title of the expected window.
            max_attempts (int): Maximum number of attempts to check for the expected window and close alerts.
            sleep_seconds (int): Time to sleep between each search for modal
            ignored_titles (list[str]): List of window titles to ignore during the search. Defaults to an empty list.
        """
        with contextlib.suppress(_ctypes.COMError, OSError):
            attempts = 0
            while attempts < max_attempts:
                current_top_title = self.dialog.child_window(control_type="Window", found_index=0).window_text()
                if current_top_title == expected_window_title:
                    break

                alerts = self.dialog.children(control_type="Window")
                for alert in alerts:
                    if alert.window_text() not in [expected_window_title] + ignored_titles:
                        logger.debug(f"Closing Alert with title: {alert.window_text()}.")
                        alert.close()

                attempts += 1
                time.sleep(sleep_seconds)
            else:
                raise ElementNotFoundError(f"Window {expected_window_title} not visible")

    def connect_to_app(self, app_path: str = None, title: str = None, timeout: int = None, **kwargs) -> Application:
        """Connect to the application using optional parameters for the app path, title, and timeout.

        This method connects to an application using the provided keyword arguments. Only the
        arguments with values will be passed to the `connect` method, allowing for flexible
        connections to applications.

        Args:
            app_path (str, optional): The location of the application on the system.
            title (str, optional): The title of the application window to connect to.
            timeout (int, optional): The maximum time (in seconds) to wait for the connection.
            **kwargs: Additional keyword arguments for connecting to the application.

        Returns:
            Application: The connected Application object.
        """
        connect_args = {
            k: v for k, v in {"path": app_path, "title": title, "timeout": timeout}.items() if v is not None
        }
        return Application(backend="uia").connect(**connect_args, **kwargs)

    @retry_if_pywin_error(exceptions=(ProcessNotFoundError,), delay=5, retries=3)  # type: ignore[misc]
    def start_app(self, app_path: str, app_folder: str, sleep_seconds: int = 2) -> None:
        """Start the App.

        Args:
            app_path (str): the location of the app on the system
            app_folder (str): the location of the folder that contains the app
            sleep_seconds (int): Time to sleep after open the app
        """
        Application().start(app_path, work_dir=app_folder)
        self.app = self.connect_to_app(path=app_path, timeout=120)
        time.sleep(sleep_seconds)

    def get_app_session_if_running(self, app_path: str) -> Union[str, None]:
        """Return current session ID if app is running.

        Args:
            app_name (str): The name of the application executable.

        Returns:
            Union[str, None]: None if the app is not running, otherwise the tasklist output.
        """
        if not CONFIG.username:
            raise UsernameNotFoundError("Username is empty for this server")
        tasklist_cmd = (
            f'tasklist /fi "imagename eq {os.path.basename(app_path)}" /fi "username eq {CONFIG.username}" /fo csv'
        )
        tasklist_result = subprocess.run(tasklist_cmd, shell=True, capture_output=True, text=True, check=True)
        tasklist_output = tasklist_result.stdout.strip()
        if "INFO: No tasks are running which match the specified criteria." in tasklist_output:
            return None
        return tasklist_output

    def kill_app(self, process: str, app_path: str) -> None:
        """Close the app if it is still open after closing the session.

        Args:
            process (str): The process name of the application.
            app_path (str): The system path of the application executable.

        Returns:
            None
        """
        app_name = os.path.basename(app_path)
        try:
            if not process:
                logger.warning(f"No running process found for {app_name}.")
                return
            if not CONFIG.username:
                raise UsernameNotFoundError("Username is empty for this server")
            process_id = next(
                int(line.split(",")[1].replace('"', "").strip()) for line in process.split("\n") if app_name in line
            )
            taskkill_cmd = f'taskkill /f /pid {process_id} /fi "username eq {CONFIG.username}"'
            subprocess.run(taskkill_cmd, shell=True, check=True)
            logger.info(f"Process {app_name} terminated for user {CONFIG.username}")
            time.sleep(2)
        except StopIteration:
            logger.warning(f"No running process found for {app_name}.")

    def get_element(
        self,
        control_type: str,
        title: Union[str, None] = None,
        auto_id: Union[str, None] = None,
        dialog: WindowSpecification = None,
        **kwargs,
    ):
        """Retrieve an element from the dialog using either 'title' or 'auto_id'.

        Args:
            control_type (str): The control type of the element (required).
            title (str, optional): The title of the element.
            auto_id (str, optional): The automation ID of the element.
            dialog (WindowSpecification, optional): The dialog to operate on. Defaults to self.dialog.

        Raises:
            ValueError: If neither 'title' nor 'auto_id' is provided.

        Returns:
            WindowSpecification: The matching element.
        """
        if dialog is None:
            dialog = self.dialog

        # Add non-None elements to kwargs
        if title is not None:
            kwargs["title"] = title
        if auto_id is not None:
            kwargs["auto_id"] = auto_id

        return dialog.child_window(control_type=control_type, **kwargs)

    def set_input_text(
        self, text: str, auto_id: str = None, title: str = None, dialog: Union[WindowSpecification, None] = None
    ) -> None:
        """Set the text of an input field in the dialog window.

        Args:
            text (str): The text to set in the input field.
            auto_id (str, optional): The automation ID of the input field. Defaults to None.
            title (str, optional): The title of the input field. Defaults to None.
            dialog (WindowSpecification, optional): The dialog to operate on. Defaults to self.dialog.

        Raises:
            ValueError: If neither 'auto_id' nor 'title' is provided.

        Note:
            This method suppresses `_ctypes.COMError` and `OSError` exceptions.
        """
        with contextlib.suppress(_ctypes.COMError, OSError):
            element = self.get_element(control_type="Edit", auto_id=auto_id, title=title, dialog=dialog)
            element.set_edit_text(text)

    def invoke_button(
        self, auto_id: str = None, title: str = None, dialog: Union[WindowSpecification, None] = None, **kargs
    ) -> None:
        """Invoke a button in the dialog window.

        Args:
            auto_id (str, optional): The automation ID of the button. Defaults to None.
            title (str, optional): The title of the button. Defaults to None.
            dialog (WindowSpecification, optional): The dialog to operate on. Defaults to self.dialog.
            **kargs: Additional keyword arguments passed to the child_window method.

        Raises:
            ValueError: If neither 'auto_id' nor 'title' is provided.
        """
        with contextlib.suppress(_ctypes.COMError, OSError):
            element = self.get_element(control_type="Button", auto_id=auto_id, title=title, dialog=dialog, **kargs)
            element.invoke()

    def wait_until_element_visible(
        self,
        control_type: str,
        title: str = None,
        auto_id: str = None,
        timeout: int = 5,
        retries=3,
        raise_ex: bool = True,
        **kwargs,
    ) -> bool:
        """Waits for an element to become visible in the dialog window.

        Args:
            control_type (str): The control type of the element (required).
            title (str, optional): The title of the element. Defaults to None.
            auto_id (str, optional): The automation ID of the element. Defaults to None.
            timeout (int, optional): The maximum time to wait for the element to become visible. Defaults to 5 seconds.
            raise_ex (bool, optional): If you wanna suppress the exception and receive boolean instead.
            **kwargs: Accept all keyword arguments that can be used to identify a child window element
        Raises:
            ValueError: If neither 'title' nor 'auto_id' is provided.
            TimeoutError: If the element does not become visible within the timeout.

        Returns:
            bool: Returns True if the element becomes visible within the allowed time.
        """
        retry = 0
        while retry < retries:
            element = self.get_element(control_type=control_type, title=title, auto_id=auto_id, **kwargs)
            try:
                element.wait("visible", timeout)
            except PywinTimeout:
                retry += 1
                continue
            return True
        if raise_ex:
            raise PywinTimeout("Element was not visible in the max timeout")
        else:
            return False

    @staticmethod  # type: ignore[misc]
    @retry_if_pywin_error(exceptions=(AssertionError,), delay=5, retries=2)  # type: ignore[misc]
    def get_element_coordinates(element: WindowSpecification) -> tuple:
        """This method gets the element coordinates.

        Args:
            element (UIAWrapper): element
        Returns:
            tuple: element coordinates
        """
        point = None
        with contextlib.suppress(_ctypes.COMError, OSError):
            point = element.rectangle().mid_point()
        if not point:
            raise AssertionError("Failed to retrieve coordinates.")
        return point

    def mouse_click_element(
        self, element: WindowSpecification, button: str = "left", offset_x: int = 0, offset_y: int = 0
    ) -> None:
        """This method performs a mouse click at a specified offset within a given element's location.

        Args:
            element (WindowSpecification): The UI element to click.
            button (str): The mouse button to click (left/right). Defaults to "left".
            offset_x (int): The horizontal offset from the element's top-left corner. Defaults to 0.
            offset_y (int): The vertical offset from the element's top-left corner. Defaults to 0.
        """
        coords = self.get_element_coordinates(element)
        click_x = coords[0] + offset_x
        click_y = coords[1] + offset_y
        element_rec = element.rectangle()

        # Ensure the point is within the bounds of the element
        click_x = min(max(click_x, element_rec.left), element_rec.right - 1)
        click_y = min(max(click_y, element_rec.top), element_rec.bottom - 1)

        mouse.click(button=button, coords=(click_x, click_y))

    def mouse_double_click_element(
        self, element: WindowSpecification, button: str = "left", set_focus: bool = False
    ) -> None:
        """This method double click with mouse in a given element's location.

        Args:
            element (UIAWrapper): element
            button (str): Mouse Button to be clicked (left/right)
            set_focus (bool): If True, calls set_focus() on the element
        """
        if set_focus:
            element.set_focus()

        mouse.double_click(button=button, coords=self.get_element_coordinates(element))

    def select_dropdown_item(self, auto_id: str, dropdown_item: str) -> None:
        """Select the required item from the dropdown list.

        Args:
            auto_id (str): the name ID the dropdown.
            dropdown_item (str): the name of the dropdown item to be selected.
        """
        with contextlib.suppress(_ctypes.COMError):
            self.dialog.child_window(auto_id=auto_id).select(dropdown_item)

    @retry_if_pywin_error()  # type: ignore[misc]
    def right_click_empty_space(self, element: WindowSpecification, sleep_seconds: int = 2) -> None:
        """This method right clicks in the blank space after a list of elements.

        Args:
            element (WindowSpecification): element
            sleep_seconds (int): time to sleep after right clicking, to wait dropdown to appear.
        """
        last_child_coords = self.get_element_coordinates(element.descendants()[-1])
        parent_rect = element.rectangle()

        # Calculate a point slightly beyond the last child but within the parent
        click_x = last_child_coords[0]
        click_y = last_child_coords[1] + 30

        # Ensure the point is within the bounds of the parent
        if click_x > parent_rect.right:
            click_x = parent_rect.right - 1
        if click_y > parent_rect.bottom:
            click_y = parent_rect.bottom - 1

        mouse.click(button="right", coords=(click_x, click_y))
        time.sleep(sleep_seconds)

    def wait_to_disappear(
        self, control_type: str, auto_id: str = None, title: str = None, max_attempts: int = 3, timeout: int = 1
    ) -> None:
        """Wait for an element to disappear and raise an AssertionError if the element is still present.

        Args:
            control_type (str): The control type of the element (required).
            auto_id (str, optional): The automation ID of the element to wait for. Defaults to None.
            title (str, optional): The title of the element to wait for. Defaults to None.
            max_attempts (int, optional): The maximum number of attempts to check. Defaults to 3.
            timeout (int, optional): The time to wait between attempts in seconds. Defaults to 1 second.

        Raises:
            ValueError: If neither 'auto_id' nor 'title' is provided, or if both are provided.
            AssertionError: If the element is still present after all attempts.
        """
        attempts = 0
        while attempts < max_attempts:
            try:
                element = self.get_element(control_type=control_type, auto_id=auto_id, title=title)
                if not element.exists():
                    return  # Element disappeared
            except (TimeoutError, ElementNotFoundError):
                return  # Element already gone or not found

            time.sleep(timeout)
            attempts += 1

        raise AssertionError(f"Element is still present after {max_attempts} attempts.")

    def click_input(
        self,
        control_type: str,
        auto_id: Union[str, None] = None,
        title: Union[str, None] = None,
        dialog: Union[WindowSpecification, None] = None,
        **kwargs,
    ) -> None:
        """Click element with based on provided kwargs.

        Args:
            control_type (str): The control type of the element (required).
            auto_id (str, optional): The automation ID of the button. Defaults to None.
            title (str, optional): The title of the button. Defaults to None.
            dialog (WindowSpecification, optional): The dialog to operate on. Defaults to self.dialog.
            **kwargs: Accept all keyword arguments that can be used to identify a child window element
        """
        with contextlib.suppress(_ctypes.COMError):
            element = self.get_element(control_type=control_type, auto_id=auto_id, title=title, dialog=dialog, **kwargs)
            element.click_input()



"""Top-level package for t-desktop."""

__author__ = """Thoughtful"""
__email__ = "support@thoughtful.ai"
__version__ = "__version__ = '0.0.11'"

from .t_desktop import DesktopApp
from .decorators import relogin_and_retry_if_pywin_error, retry_if_pywin_error, capture_screenshot_if_pywin_error

__all__ = [
    "DesktopApp",
    "relogin_and_retry_if_pywin_error",
    "retry_if_pywin_error",
    "capture_screenshot_if_pywin_error",
]



import os
import time
from datetime import datetime
from functools import wraps
from pathlib import Path
from typing import Callable, List, Type
from t_bug_catcher import attach_file_to_exception
from PIL import ImageGrab


from .config import CONFIG, IS_WINDOWS_OS
from .exceptions import AppCrashException, LoginMethodNotFoundError
from .utils.logger import logger

if IS_WINDOWS_OS:
    from pywinauto.application import ProcessNotFoundError
    from pywinauto.findwindows import ElementAmbiguousError, ElementNotFoundError, WindowNotFoundError
    from pywinauto.timings import TimeoutError
else:
    logger.warning("Not in a Windows Machine, failed to import UI related libraries")


def relogin_and_retry_if_pywin_error(retries: int = 3, delay: int = 1) -> callable:
    """Decorator to relogin and retry if pywin error.

    Args:
        retries (int): Number of times to retry the function.
        delay (int): Delay in seconds between retries.

    Returns:
        Callable: The decorated function with retry logic.
    """
    if retries < 0:
        raise ValueError("Retries must be a non-negative integer.")

    default_exceptions = (
        ElementNotFoundError,
        WindowNotFoundError,
        TimeoutError,
        AppCrashException,
        OSError,
        RuntimeError,
        ProcessNotFoundError,
    )

    def decorator(func: callable) -> callable:
        @wraps(func)
        def wrapper(self: object, *args, **kwargs) -> callable:
            if not hasattr(self, "login") or not callable(getattr(self, "login")):
                raise LoginMethodNotFoundError("The 'login' method is not defined or not callable.")

            for attempt in range(retries + 1):
                try:
                    return func(self, *args, **kwargs)
                except default_exceptions as e:
                    logger.warning(f"Failed to process due to crash: {str(e)}. Retry attempt: {attempt + 1}")
                    if attempt < retries:
                        logger.info("Attempting to re-login.")
                        self.login()
                        time.sleep(delay)
                    else:
                        raise e

        return wrapper

    return decorator


def retry_if_pywin_error(
    retries: int = 3,
    delay: int = 1,
    close_modal: bool = False,
    exceptions: tuple[BaseException] = (),
) -> callable:
    """Decorator to be used in Apps derived classes to retry a function if specified exceptions occur.

    Args:
        retries (int): Number of times to retry the function.
        delay (int): Delay in seconds between retries.
        close_modal (bool): Whether to attempt closing modals on error.
        exceptions (Tuple[Type[BaseException], ...]): Tuple of exceptions to catch and retry on.

    Returns:
        Callable: The decorated function with retry logic.
    """
    # Default exceptions
    default_exceptions = (
        ElementNotFoundError,
        WindowNotFoundError,
        TimeoutError,
        AppCrashException,
        ProcessNotFoundError,
        RuntimeError,
    )
    if retries < 0:
        raise ValueError("Retries must be a non-negative integer.")

    exceptions = exceptions if isinstance(exceptions, tuple) else (exceptions,)
    all_exceptions = default_exceptions + exceptions

    def decorator(func: callable) -> callable:
        @wraps(func)
        def wrapper(self: object, *args, **kwargs) -> callable:
            last_exception = None
            for attempt in range(retries + 1):
                try:
                    result = func(self, *args, **kwargs)
                    return result
                except all_exceptions as e:
                    last_exception = e
                    logger.warning(f"Failed to process due to error: {str(e)}. Retry attempt: {attempt + 1}")
                    if close_modal:
                        if hasattr(self, "close_modal"):
                            self.close_modal()
                        else:
                            self.app.top_window().close()
                    time.sleep(delay)

            if last_exception is not None:
                raise last_exception

        return wrapper

    return decorator


def capture_screenshot_if_pywin_error(
    func: Callable = None,  # Func as default argument to allow use without parenthesis
    exceptions_to_include: List[Type[Exception]] = None,
    output: str = CONFIG.DIRECTORIES.SCREENSHOTS,
) -> Callable:
    """Decorator to capture a screenshot if specific exceptions occur.

    This decorator wraps a function and captures a screenshot when one of the
    specified exceptions is raised. The screenshot will be saved to the output
    directory, and if `ImageGrab` fails, the `RPA.Desktop` module will be used
    as a fallback (if installed).

    Args:
        func (Callable, optional): The function being decorated.
        exceptions_to_include (List[Type[Exception]], optional): A list of exception
            types to trigger the screenshot capture. Defaults to None.
        output (str, optional): Directory path where screenshots will be saved.
            Defaults to `CONFIG.DIRECTORIES.SCREENSHOTS`.

    Returns:
        Callable: The decorated function that captures a screenshot upon specific errors.

    Raises:
        Exception: Re-raises the original exception that triggered the screenshot.
    """
    default_exceptions_to_include = [
        AppCrashException,
        TimeoutError,
        ElementNotFoundError,
        WindowNotFoundError,
        AttributeError,
        IndexError,
        RuntimeError,
        ElementAmbiguousError,
    ]

    def decorator(func: Callable) -> Callable:
        exceptions_to_include_list = default_exceptions_to_include + (exceptions_to_include or [])

        @wraps(func)
        def wrapper(*args, **kwargs) -> Callable:
            try:
                return func(*args, **kwargs)
            except tuple(exceptions_to_include_list) as exception:
                ts = datetime.now().strftime("%H_%M_%S")
                file_name = f"{ts}_{func.__name__}.png"
                Path(output).mkdir(parents=True, exist_ok=True)
                file_path = os.path.join(output, file_name)

                try:
                    # Try to capture the screenshot
                    screenshot = ImageGrab.grab()
                    screenshot.save(file_path)
                    logger.info(f"Screenshot saved to {file_path}")
                    attach_file_to_exception(exception, file_path)
                except Exception as e:
                    try:
                        from RPA.Desktop import Desktop as RpaDesktop

                        logger.warning(f"Failed to capture screenshot using ImageGrab. Cause: {e}")
                        desktop = RpaDesktop()
                        desktop.take_screenshot(file_path)
                        logger.info(f"Screenshot saved to {file_path}")
                        attach_file_to_exception(exception, file_path)
                    except ImportError:
                        logger.warning("rpaframework not installed")
                    raise exception
                raise exception

        return wrapper

    if func is not None:
        return decorator(func)

    return decorator



from RPA.Robocorp.WorkItems import WorkItems
from urllib.parse import urlparse
import os

LOCAL_RUN = os.environ.get("RC_PROCESS_RUN_ID") is None

if not LOCAL_RUN:
    work_items = WorkItems()
    work_items.get_input_work_item()
    work_item = work_items.get_work_item_variables()
    fabric_variables = work_item.get("variables", dict())
    metadata = work_item.get("metadata", {})
    run_number = metadata.get("processRunUrl", "").split("/")[-1]

    rc_run_link = (
        f"https://cloud.robocorp.com/organizations/{os.environ.get('RC_ORGANIZATION_ID', '')}"
        f"/workspaces/{os.environ.get('RC_WORKSPACE_ID')}/processes"
        f"/{os.environ.get('RC_PROCESS_ID')}/runs/{os.environ.get('RC_PROCESS_RUN_ID', '')}/"
        f"stepRuns/{os.environ.get('RC_ACTIVITY_RUN_ID')}/"
    )
    empower_url = metadata.get("process", dict()).get("processRunUrl", "") or fabric_variables.get("processRunUrl", "")
    empower_url_text = urlparse(empower_url).path.split("/")[-1]

else:
    rc_run_link = "https://cloud.robocorp.com"
    empower_url = "https://app.thoughtful-dev.ai"
    empower_url_text = "Empower"

OUTPUT_FOLDER = os.path.join(os.getcwd(), "output")



"""Deals with memory related stuffs."""
import psutil


class MemoryManager:
    """A class to manage and retrieve memory usage information."""

    @staticmethod
    def get_memory_stats() -> tuple[float, float, float, float]:
        """Retrieves memory statistics for the current user, total system, system processes, and available memory.

        Returns:
            tuple: A tuple containing:
                - Total RAM used by the current user in gigabytes (GB) with two decimal places.
                - Total RAM used by all users in gigabytes (GB) with two decimal places.
                - Average RAM used by system processes in gigabytes (GB) with two decimal places.
                - Available RAM in gigabytes (GB) with two decimal places.

        Raises:
            None
        """
        current_user_memory = MemoryManager._get_current_user_memory_usage()
        total_memory_used = MemoryManager._get_total_memory_usage()
        system_processes_memory = MemoryManager._get_system_processes_memory_usage()
        available_memory = MemoryManager._get_available_memory()

        return (
            round(current_user_memory, 2),
            round(total_memory_used, 2),
            round(system_processes_memory, 2),
            round(available_memory, 2),
        )

    @staticmethod
    def _get_current_user_memory_usage() -> float:
        """Retrieves the total RAM used by the current user.

        Returns:
            float: RAM used by the current user in gigabytes (GB).

        Raises:
            None
        """
        current_user = psutil.Process().username()  # Get the current user's username
        total_memory = 0.0

        for proc in psutil.process_iter(["memory_info", "username"]):
            try:
                if proc.info["username"] == current_user:
                    total_memory += proc.info["memory_info"].rss
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                continue

        return total_memory / (1024**3)  # Convert from bytes to GB

    @staticmethod
    def _get_total_memory_usage() -> float:
        """Retrieves the total RAM used by all users.

        Returns:
            float: Total RAM used in gigabytes (GB).

        Raises:
            None
        """
        total_memory = sum(proc.memory_info().rss for proc in psutil.process_iter() if proc.is_running())
        return total_memory / (1024**3)  # Convert from bytes to GB

    @staticmethod
    def _get_system_processes_memory_usage() -> float:
        """Retrieves the total RAM used by system processes, excluding users with 'thoughtful' in their names.

        Returns:
            float: Total RAM used by system processes in gigabytes (GB), rounded to 2 decimal places.

        Raises:
            None
        """
        system_memory = 0.0

        for proc in psutil.process_iter(["memory_info", "username"]):
            try:
                username = proc.info.get("username", "")
                if username and "thoughtful" in username.lower():
                    continue
                system_memory += proc.info["memory_info"].rss / (1024**3)  # Convert bytes to GB
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                continue

        return round(system_memory, 2)

    @staticmethod
    def _get_available_memory() -> float:
        """Retrieves the available system RAM.

        Returns:
            float: Available RAM in gigabytes (GB).

        Raises:
            None
        """
        available_memory = psutil.virtual_memory().available
        return available_memory / (1024**3)  # Convert from bytes to GB

    @staticmethod
    def get_total_ram() -> float:
        """Retrieves the total RAM in the machine.

        Returns:
            float: Total RAM in the machine in gigabytes (GB) with two decimal places.

        Raises:
            None
        """
        total_ram = psutil.virtual_memory().total
        return round(total_ram / (1024**3), 2)  # Convert from bytes to GB



"""Monitor Server."""
import datetime
import os
import threading
import time
from openpyxl import load_workbook
import pytz
import atexit
from typing import Any, Optional
from t_desktop.utils.logger import logger
from t_desktop.monitor.workitems import rc_run_link, empower_url, empower_url_text, OUTPUT_FOLDER
from t_desktop.monitor.managers.cpu_manager import CPUManager
from t_desktop.monitor.managers.disk_manager import DiskManager
from t_desktop.monitor.managers.memory_manager import MemoryManager
from t_desktop.monitor.spreadhseet.g_sheets import GSheets
from google.oauth2.service_account import Credentials


class ServerMonitor:
    """A class to manage and monitor system performance in a separate thread."""

    _instance = None

    def __new__(cls, *args, **kwargs):
        """Ensure only one instance of the class is created (Singleton Pattern)."""
        if cls._instance is None:
            cls._instance = super(ServerMonitor, cls).__new__(cls)
        return cls._instance

    def __init__(
        self,
        process_name: str,
        hostname: str,
        interval: int = 60,
        export_path: Optional[str] = OUTPUT_FOLDER,
    ) -> None:
        """Initialize the server manager with configurations.

        Args:
            process_name (str): The process name to monitor.
            hostname (str): The hostname for the server.
            spreadsheet_id (str, optional): The ID of the Google Spreadsheet.
            spreadsheet_tab (str, optional): The tab name or index within the spreadsheet.
            google_sheet_credentials (list, optional): Credentials Object required for Google Sheets access.
            interval (int, optional): Interval in seconds to tick the stats. Defaults to 60.
            export_path (str, optional): Path to export results locally. Defaults to output.
            if export_path is None or empty, it will not be locally exported.
        """
        if hasattr(self, "_initialized") and self._initialized:
            return
        self._initialized = True
        self.process_name = process_name
        self._hostname = hostname
        self.interval = interval
        self.export_path = export_path

        self._initialize_managers()
        self._initialize_metrics()
        self._spreadsheet_id = None
        self._spreadsheet_tab = None
        self._setup_urls()
        self.g_sheets = None
        atexit.register(self.stop)

    def _initialize_managers(self) -> None:
        """Initialize resource management classes."""
        self._memory_manager = MemoryManager()
        self._cpu_manager = CPUManager()
        self._disk_manager = DiskManager()

    def _initialize_metrics(self) -> None:
        """Initialize monitoring metrics."""
        self._running = False
        self._thread = None
        self._cpu_usage = []
        self._memory_usage = []
        self._peak_memory_usage = 0.0
        self._peak_cpu_usage = 0.0
        self._system_cpu_usage = []
        self._peak_system_cpu_usage = 0.0
        self._all_users_memory_usage = []
        self._system_process_memory = []
        self._user_cpu_usage = []
        self._start_time = None

    def _setup_urls(self) -> None:
        """Setup URLs for monitoring links."""
        self._empower_url = f'=HYPERLINK("{empower_url}", "{empower_url_text}")'
        self._rc_run_link = f'=HYPERLINK("{rc_run_link}", "RC RUN LINK")'
        self._env = "DEV" if "dev" in empower_url else "PROD"

    def start(self) -> None:
        """Start the monitoring thread."""
        if self._running:
            logger.warning("ServerManager is already running.")
            return
        logger.info("Starting Server Monitor")

        self._running = True
        self._reset_metrics()
        self._thread = threading.Thread(target=self._monitor)
        self._thread.start()

    def stop(self) -> None:
        """Stop the monitoring thread and export results."""
        if not self._running:
            logger.warning("ServerManager is not running.")
            return
        logger.info("Stopping Server Monitor")

        self._running = False
        self._thread.join()

        row_data = self._build_row()
        self._export_results(row_data)

    def export_results(self):
        """Export results without stopping the monitor."""
        row_data = self._build_row()
        self._export_results(row_data)

    def _reset_metrics(self) -> None:
        """Reset monitoring metrics."""
        mst_tz = pytz.timezone("America/Denver")
        now = datetime.datetime.now(mst_tz)
        self._start_time = now.strftime("%H:%M:%S")
        self._cpu_usage.clear()
        self._memory_usage.clear()
        self._all_users_memory_usage.clear()
        self._peak_memory_usage = 0.0
        self._peak_cpu_usage = 0.0
        self._system_cpu_usage.clear()
        self._system_process_memory.clear()
        self._user_cpu_usage.clear()

    def _monitor(self) -> None:
        """Monitor system performance metrics."""
        previous_cpu_usage = None
        previous_memory_usage = None

        while self._running:
            self._collect_metrics(previous_cpu_usage, previous_memory_usage)
            time.sleep(self.interval)

    def _collect_metrics(self, previous_cpu_usage: Optional[float], previous_memory_usage: Optional[float]) -> None:
        """Collect and analyze metrics for CPU and memory usage.

        Args:
            previous_cpu_usage (Optional[float]): The CPU usage from the previous monitoring interval.
            previous_memory_usage (Optional[float]): The memory usage from the previous monitoring interval.
        """
        overall_cpu_usage, system_cpu_usage, user_cpu_usage = self._cpu_manager.get_cpu_stats()
        memory_stats = self._memory_manager.get_memory_stats()

        self._analyze_variation("CPU", previous_cpu_usage, overall_cpu_usage, 20)
        self._analyze_variation("Memory", previous_memory_usage, memory_stats[0], 20)

        self._update_metrics(overall_cpu_usage, system_cpu_usage, user_cpu_usage, memory_stats)
        previous_cpu_usage, previous_memory_usage = overall_cpu_usage, memory_stats[0]

    def _analyze_variation(
        self, metric_name: str, previous_value: Optional[float], current_value: float, threshold: float
    ) -> None:
        """Log warnings if a metric's variation exceeds a specified threshold.

        Args:
            metric_name (str): The name of the metric being analyzed (e.g., "CPU", "Memory").
            previous_value (Optional[float]): The previous value of the metric.
            current_value (float): The current value of the metric.
            threshold (float): The threshold percentage for triggering a warning.
        """
        if previous_value is not None:
            increase = current_value - previous_value
            if increase > threshold:
                logger.warning(
                    f"{metric_name} usage increased by more than {threshold}%. "
                    f"Previous: {previous_value}%, Current: {current_value}%"
                )

    def _update_metrics(
        self,
        overall_cpu_usage: float,
        system_cpu_usage: float,
        user_cpu_usage: float,
        memory_stats: tuple[float, float, float],
    ) -> None:
        """Update monitoring metrics and track peaks.

        Args:
            overall_cpu_usage (float): The total CPU usage percentage.
            system_cpu_usage (float): The system CPU usage percentage.
            user_cpu_usage (float): The user CPU usage percentage.
            memory_stats (Tuple[float, float, float]): A tuple containing:
                - User memory usage (in GB).
                - Memory usage by all users (in GB).
                - System memory usage (in GB).
        """
        self._cpu_usage.append(overall_cpu_usage)
        self._system_cpu_usage.append(system_cpu_usage)
        self._user_cpu_usage.append(user_cpu_usage)

        current_user_memory_usage = memory_stats[0]
        all_users_memory_usage = memory_stats[1]

        self._memory_usage.append(current_user_memory_usage)
        self._all_users_memory_usage.append(all_users_memory_usage)
        self._system_process_memory.append(memory_stats[2])

        self._peak_memory_usage = max(self._peak_memory_usage, current_user_memory_usage)
        self._peak_cpu_usage = max(self._peak_cpu_usage, overall_cpu_usage)

    def _export_results(self, row_data: list[Any]) -> None:
        """Export collected results to a spreadsheet.

        Args:
            row_data (List[Any]): The data to export as a row.
        """
        if self.export_path:
            self._export_to_local_file(row_data)
        if self.g_sheets:
            self._upload_to_google_sheets(row_data)

    def _upload_to_google_sheets(self, row_data: list[Any]) -> None:
        """Upload results to Google Sheets.

        Args:
            row_data (List[Any]): The data to export as a row.
        """
        logger.info("Uploading results to spreadsheet")
        self.g_sheets.append_row(self._spreadsheet_id, self._spreadsheet_tab, row_data)

    def _export_to_local_file(self, row_data: list[Any]) -> None:
        """Export results to a local file.

        Args:
            row_data (List[Any]): The data to export as a row.
        """
        try:
            export_path = self._determine_export_path()
            workbook = load_workbook(self._template_path())
            sheet = workbook.active
            sheet.append(row_data)
            workbook.save(export_path)
            logger.info("Results exported to %s", export_path)
        except Exception as e:
            logger.error("Failed to append to and export spreadsheet: %s", e)

    def setup_google_sheets(
        self,
        credentials: Credentials,
        spreadsheet_id: str = None,
        spreadsheet_tab: str = None,
    ) -> None:
        """Setup the google sheets with the given credentials.

        Args:
            credentials (Credentials): The google sheets credentials object.
            spreadsheet_id (str): spreadsheet_id
            spreadsheet_tab (str): spreadsheet_tab
        """
        self._spreadsheet_id = spreadsheet_id
        self._spreadsheet_tab = spreadsheet_tab
        self.g_sheets = GSheets(credentials=credentials)

    def _determine_export_path(self) -> str:
        """Determine export path for the spreadsheet.

        If `self.export_path` is not specified, use the 'output' folder in the
        project root and create it if necessary.

        Returns:
            str: The full export path for the spreadsheet.
        """
        os.makedirs(self.export_path, exist_ok=True)
        return os.path.join(self.export_path, "PerformanceMonitor.xlsx")

    def _template_path(self) -> str:
        """Get path to the local spreadsheet template."""
        script_dir = os.path.dirname(os.path.abspath(__file__))
        return os.path.join(script_dir, "assets", "Template.xlsx")

    def _build_row(self) -> list:
        """Build the row with the result values."""
        average_cpu_usage = sum(self._cpu_usage) / len(self._cpu_usage) if self._cpu_usage else 0
        average_memory_usage = sum(self._memory_usage) / len(self._memory_usage) if self._memory_usage else 0
        average_all_users_memory_usage = (
            sum(self._all_users_memory_usage) / len(self._all_users_memory_usage) if self._all_users_memory_usage else 0
        )
        average_system_memory_usage = (
            sum(self._system_process_memory) / len(self._system_process_memory) if self._system_process_memory else 0
        )
        average_system_cpu_usage = (
            sum(self._system_cpu_usage) / len(self._system_cpu_usage) if self._system_cpu_usage else 0
        )
        average_user_cpu_usage = sum(self._user_cpu_usage) / len(self._user_cpu_usage) if self._user_cpu_usage else 0
        disk_status = self._disk_manager.get_disk_stats()

        total_ram = self._memory_manager.get_total_ram()

        # Calculating percentages
        user_memory_usage_percent = (average_memory_usage / total_ram) * 100 if total_ram else 0
        all_users_memory_usage_percent = (average_all_users_memory_usage / total_ram) * 100 if total_ram else 0
        system_memory_usage_percent = (average_system_memory_usage / total_ram) * 100 if total_ram else 0
        peak_memory_usage_percent = (self._peak_memory_usage / total_ram) * 100 if total_ram else 0

        mst_tz = pytz.timezone("America/Denver")
        now = datetime.datetime.now(mst_tz)

        # Format the date and time separately
        date = now.strftime("%Y-%m-%d")
        end_time = now.strftime("%H:%M:%S")

        # Create row_data with separate date and time
        return [
            date,  # Date
            self._start_time,  # Start Time
            end_time,  # End Time
            self._env,  # Empower ENV
            self._empower_url,  # Empower Run ID
            self._rc_run_link,  # RC Step Run Link
            self.process_name,  # RC Process Name
            self._hostname,  # Server Name
            os.getenv("USERNAME"),  # Username
            f"{average_user_cpu_usage:.2f}",  # [User] AVG CPU Usage
            f"{average_cpu_usage:.2f}",  # [Total] AVG CPU Usage
            f"{self._peak_cpu_usage:.2f}",  # [User] Peak CPU Usage
            f"{average_system_cpu_usage:.2f}",  # System CPU Usage
            f"{average_memory_usage:.2f}",  # [User] Avg Ram Usage GB
            f"{user_memory_usage_percent:.2f}%",  # [User] Avg Ram Usage %
            f"{self._peak_memory_usage:.2f}",  # [User] Peak Ram Usage GB
            f"{peak_memory_usage_percent:.2f}%",  # [User] Peak Ram Usage %
            f"{average_all_users_memory_usage:.2f}",  # [All Users] Avg Ram Usage GB
            f"{all_users_memory_usage_percent:.2f}%",  # [All Users] Avg Ram Usage %
            f"{average_system_memory_usage:.2f}",  # [System] Avg Ram Usage GB
            f"{system_memory_usage_percent:.2f}%",  # [System] Avg Ram Usage %
            f"{total_ram:.2f}",  # Total RAM GB
            f"{disk_status[0]:.2f}",  # Total Disk Space
            f"{disk_status[1]:.2f}",  # Available Disk Space
            f"{disk_status[2]:.2f}",  # Used Disk Space
        ]
